local array = include( "modules/array" )
local util = include( "modules/util" )
local cdefs = include( "client_defs" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local simfactory = include("sim/simfactory")
local unitdefs = include("sim/unitdefs")
local abilityutil = include( "sim/abilities/abilityutil" )

local brokenDoor = {
	door = true,
	closed = true,
	locked = nil,
	openSound = simdefs.SOUND_DOOR_OPEN,
	closeSound = simdefs.SOUND_DOOR_CLOSE,
	breakSound = simdefs.SOUND_DOOR_BREAK,
	keybits =  simdefs.DOOR_KEYS.OFFICE,
	wallUVs = {
		unlocked = cdefs.WALL_DOOR_UNLOCKED,
		locked = cdefs.WALL_DOOR_LOCKED,
		broken = cdefs.WALL_DOOR_BROKEN,
	},
}

local tileDir = {
	[0] = { 1, 0 },
	[2] = { 0, 1 },
	[4] = {-1, 0 },
	[6] = { 0,-1 },
}

local function toExitID( cellID, dir )
	return cellID * 4 + (dir / 2)
end

local SOUND_BREACH = {
	path = "urist/Actions/wallbreacher",
	range = 2, --simdefs.SOUND_RANGE_2, --was exploitably far
}

return {
	name = STRINGS.URIST.ACTION_BREACH,

	getName = function( self, sim, abilityOwner, abilityUser )
		return self.name
	end,

	createToolTip = function( self, sim, unit )
		local desc = STRINGS.URIST.ACTION_BREACH_TIP
		if false then --TODO if damaged
			desc = STRINGS.URIST.ACTION_BREACH_RESUME_TIP
		end
		return abilityutil.formatToolTip( STRINGS.URIST.ACTION_BREACH, desc, simdefs.DEFAULT_COST )
	end,

	profile_icon = "gui/icons/item_icons/items_icon_small/icon-item_wallbreacher_small.png",

	proxy = true,

	acquireTargets = function( self, targets, game, sim, abilityOwner, abilityUser )
		local exits = {}
		local x0, y0 = abilityUser:getLocation()
		local fromCell = sim:getCell( x0, y0 )
		local player = abilityUser:getPlayerOwner()

		for dir, side in pairs( fromCell.sides ) do
			local rdir = simquery.getReverseDirection(dir)
			local x1, y1 = x0 + tileDir[dir][1], y0 + tileDir[dir][2]
			local rCell = sim:getCell( x1, y1 )

			if simquery.checkIsWall( sim, fromCell, dir )
			and rCell and not rCell.isSolid
			and player:getLastKnownCell( sim, x1, y1 ) then
				table.insert( exits, { x = x0, y = y0, dir = dir } )
			end
		end

		return targets.exitTarget( game, exits, self, abilityOwner, abilityUser )
	end,

	canUseAbility = function( self, sim, abilityOwner, abilityUser, x, y, dir )
		-- Must be the owner
		if abilityOwner ~= abilityUser and abilityOwner:getUnitOwner() ~= abilityUser then
			return false
		end

		-- A software developer may call this DRY-wall. Badumts
		if self.target and self.target.x == x and self.target.y == y and self.target.dir == dir then
			return false, STRINGS.UI.REASON.ALREADY_EQUIPPED
		end

		if rawget(_G,"SCRIPT_PATHS") == nil or not SCRIPT_PATHS.W93_FuncLibActive then
			return false, STRINGS.URIST.FUNCLIB_MISSING
		end

		return abilityutil.checkRequirements( abilityOwner, abilityUser )
	end,

	executeAbility = function( self, sim, abilityOwner, abilityUser, x0, y0, dir )
		local fromCell = sim:getCell( x0, y0 )
		assert( fromCell and dir )

		abilityUser:resetAllAiming()
		-- sim:dispatchEvent( simdefs.EV_UNIT_USEDOOR, { unitID = abilityUser:getID(), facing = dir } )	

		-- log:write("executeAbility ".. tostring(x0) .."/".. tostring(y0) .." :".. tostring(dir))
		self.target = { x = x0, y = y0, dir = dir }

		abilityUser:setFacing(dir)
		-- Hack (for visuals)
		abilityUser:getTraits().data_hacking = abilityOwner:getID()
		abilityUser:getTraits().wallbreacher = true
		abilityOwner:getTraits().hacker = abilityUser:getID()
		abilityOwner:getTraits().wallbreacher = true
		sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = abilityUser })

		sim:processReactions( abilityUser )
		-- sim:dispatchEvent( simdefs.EV_UNIT_USEDOOR_PST, { unitID = abilityUser:getID(), facing = dir } )
	end,

	-- not a standard callback; called by wrapped methods
	stopUsingAbility = function( self, sim, abilityOwner, abilityUser )
		self.target = nil
		abilityOwner = abilityOwner or self.abilityOwner
		if abilityOwner then
			abilityUser = abilityUser or sim:getUnit( abilityOwner:getTraits().hacker )
			abilityOwner:getTraits().hacker = nil
			abilityOwner:getTraits().wallbreacher = nil
		end
		if abilityUser then
			abilityUser:getTraits().data_hacking = nil
			abilityUser:getTraits().wallbreacher = nil
			sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = abilityUser })
		end
	end,

	onSpawnAbility = function( self, sim, unit )
        self.abilityOwner = unit
		self.target = nil
		self.damage = {}
        sim:addTrigger( simdefs.TRG_START_TURN, self )
        sim:addTrigger( simdefs.TRG_UNIT_NEWTARGET, self )
	end,

	onDespawnAbility = function( self, sim, unit )
        sim:removeTrigger( simdefs.TRG_START_TURN, self )
        sim:removeTrigger( simdefs.TRG_UNIT_NEWTARGET, self )
        self.abilityOwner = nil
		self.target = nil
		self.damage = nil
	end,

    onTrigger = function( self, sim, evType, evData )
        if evType == simdefs.TRG_START_TURN and evData:isPC() and self.target and not self.abilityOwner:isDown() then
			local x0, y0 = self.target.x, self.target.y
			local cell = sim:getCell( x0, y0 )
			local dir = self.target.dir
			local rdir = simquery.getReverseDirection(dir)
			local rCell = sim:getCell( x0 + tileDir[dir][1], y0 + tileDir[dir][2] )
			assert( cell, "Wallbreacher no longer has from cell!" )
			assert( rCell, "Wallbreacher no longer has cell on other side!" )
			assert( not rCell.isSolid, "Wallbreacher other side is suddenly solid!" )
			-- log:write("onTrigger ".. tostring(x0) .."/".. tostring(y0) .." :".. tostring(dir))
			if not simquery.checkIsWall( sim, cell, dir ) then
				log:write("WARNING: Wall vanished while breaching in progress")
				self:stopUsingAbility( sim )
				return
			end

			-- local wall_dmg = (cell.sides[dir].damage or 0)
			local wall_dmg = (self.damage[toExitID( cell.id, dir )] or 0)
			if wall_dmg < 2 then
				log:write("Damaging Wall (".. tostring(wall_dmg) ..") from ".. tostring(x0) ..",".. tostring(y0) .." (id ".. tostring(cell.id) ..") facing ".. tostring(dir) )
				-- Increment damage
				self.damage[toExitID( cell.id, dir )] = wall_dmg + 1
				self.damage[toExitID( rCell.id, rdir )] = wall_dmg + 1
				-- This does not reset with rewinds.
				-- cell.sides[dir].damage = wall_dmg + 1
				-- rCell.sides[rdir].damage = wall_dmg + 1
				sim:emitSound( SOUND_BREACH, x0, y0, self.abilityOwner)
				sim:processReactions( self.abilityOwner )
			else
				log:write("Breaching Wall using ".. tostring(self.abilityOwner:getName()))
				-- The following is copied from the Function Library simemppack (Breaching Charges)
				-- I, Mobbstar, have no idea why these dupes happen
				cell.sides = util.tdupe(cell.sides)
				rCell.sides = util.tdupe(rCell.sides)

				cell.exits = util.tdupe(cell.exits)
				rCell.exits = util.tdupe(rCell.exits)

				cell.sides[dir] = util.tcopy(brokenDoor)
				rCell.sides[rdir] = util.tcopy(brokenDoor)
				cell.exits[dir] = util.tcopy(brokenDoor)
				rCell.exits[rdir] = util.tcopy(brokenDoor)
				rCell.exits[rdir].cell = cell
				cell.exits[dir].cell = rCell
				sim:dispatchEvent( "EV_WALL_DESTROYED", { cell = cell, rCell = rCell, dir = dir, rdir = rdir } )
				sim:emitSound( SOUND_BREACH, x0, y0, self.abilityOwner)
				sim:modifyExit(cell, dir, simdefs.EXITOP_BREAK_DOOR, self.abilityOwner, false)

				self:stopUsingAbility( sim )
			end
        elseif evType == simdefs.TRG_UNIT_NEWTARGET and evData.target == self.abilityOwner then
			-- when spotted, stop (standard set by Leveragents Piker)
			self:stopUsingAbility( sim )
		end
    end,
}